/*
 * linux/net/sunrpc/auth_des.c
 *
 * DES authentication for Sun RPC.
 *
 * Copyright (C) 1996, Alexander N. Kuznetsiov <kuznet@inr.ac.ru>
 *		       Olaf Kirch <okir@monad.swb.de>
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/malloc.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/utsname.h>
#include <linux/des.h>

#include <linux/sunrpc/clnt.h>
#include <linux/sunrpc/auth.h>
#include <linux/sunrpc/keycall.h>


#define DES_CRED_EXPIRE		(360 * HZ)
#define DES_MIN_TTL		(10 * HZ)
#define DES_MAXDOMAINNAME	(RPC_MAXNETNAMELEN-sizeof("unix.@")-16)

#define ADN_FULLNAME		0
#define ADN_NICKNAME		1

/*
 * DES writeslack:
 * fullname cred: flavor+len+namekind+netnamelen+netname+cryptkey+window
 * nickname cred: flavor+len+namekind+nickname
 * verifier:	  flavor+len+timestamp+winverf
 * DES readslack:
 * verifier:	  flavor+len+timestamp+nickname
 */
#define DES_WRITESLACK		(12+RPC_MAXNETNAMELEN)
#define DES_READSLACK		(5)

struct des_auth {
	struct rpc_auth		ds_base;

	char			ds_domainname[DES_MAXDOMAINNAME+1];

	unsigned long		ds_jiffies;	/* last xmit time */
	u32			ds_time[2];	/* sec/msec stamp */
	u32			ds_window;	/* validity of window */
};

struct des_cred {
	struct rpc_cred		dc_base;

	unsigned long		dc_nxtrfr;	/* next refresh */
	struct rpc_wait_queue	dc_wait;

	u32			dc_namekind;	/* full/nickname */
	u32			dc_nickname;	/* nickname */
	des_cblock		dc_key;		/* conversation key */
	des_cblock		dc_xkey;	/* encrypted conv key */
	des_key_schedule	dc_keysched;

	struct cryptkeyarg	dc_cryptarg;
	struct cryptkeyres	dc_cryptres;
};
#define dc_uid			dc_base.cr_uid
#define dc_next			dc_base.cr_next
#define dc_count		dc_base.cr_count
#define dc_flags		dc_base.cr_flags
#define dc_expire		dc_base.cr_expire


#ifdef RPC_DEBUG
# define RPCDBG_FACILITY	RPCDBG_AUTH
#endif

extern struct rpc_authops	authdes_ops;

static void	des_refresh_done(struct rpc_task *task);

/*
 * Create a DES authenticator
 */
static struct rpc_auth *
des_create(struct rpc_clnt *clnt)
{
	struct des_auth	*auth;

	dprintk("RPC: creating DES authenticator for client %p\n", clnt);
	if (!(auth = (struct des_auth *) rpc_allocate(0, sizeof(*auth))))
		return NULL;
	auth->ds_base.au_cslack = DES_WRITESLACK;
	auth->ds_base.au_rslack = DES_READSLACK;
	auth->ds_base.au_expire = DES_CRED_EXPIRE;
	strncpy(auth->ds_domainname, system_utsname.domainname,
				  DES_MAXDOMAINNAME);
	auth->ds_base.au_ops    = &authdes_ops;
	auth->ds_window         = 1800*HZ;	/* default 30 minutes */

	rpcauth_init_credcache((struct rpc_auth *) auth);

	MOD_INC_USE_COUNT;
	return (struct rpc_auth *) auth;
}

static void
des_destroy(struct rpc_auth *auth)
{
	dprintk("RPC: destroying DES authenticator %p\n", auth);
	rpcauth_free_credcache(auth);
	rpc_free(auth);
	MOD_DEC_USE_COUNT;
}

static struct des_cred *
des_new_cred(struct rpc_task * task, uid_t uid)
{
	struct des_cred	*cred;

	dprintk("RPC:      allocating DES cred for uid %d\n", uid);

	if (!(cred = (struct des_cred *) rpc_malloc(task, sizeof(*cred))))
		return NULL;
	memset(cred, 0, sizeof(*cred));
	cred->dc_count  = 0;
	cred->dc_flags  = 0;
	cred->dc_uid    = uid;
	cred->dc_nxtrfr = jiffies + DES_MIN_TTL;
	cred->dc_wait   = RPC_INIT_WAITQ("des_refresh");

	return cred;
}

static struct rpc_cred *
des_create_cred(struct rpc_task *task)
{
	return (struct rpc_cred *) des_new_cred(task,
			RPC_DO_ROOTOVERRIDE(task)? 0 : current->uid);
}

/*
 * Delete cred handle
 */
static void
des_destroy_cred(struct rpc_cred *cred)
{
	rpc_free(cred);
}

/*
 * Match a cred handle against the current task's creds
 */
static int
des_match_cred(struct rpc_task *task, struct rpc_cred *rcred)
{
	struct des_cred	*cred = (struct des_cred *) rcred;

	return cred->dc_uid == (RPC_DO_ROOTOVERRIDE(task)? 0 : current->uid);
}

/*
 * Marshal DES creds
 */
static u32 *
des_marshal(struct rpc_task *task, u32 *p, int ruid)
{
	struct des_auth	*auth = (struct des_auth *) task->tk_auth;
	struct des_cred	*cred = (struct des_cred *) task->tk_cred;
	struct rpc_rqst	*req = task->tk_rqstp;
	static u32	ivec[2] = {0, 0};
	u32		cryptbuf[4], *hold;
	int		status;

	/* Get the current time. Make sure no time stamp is used twice
	 * in order to thwart replay attempts.
	 * We use a granularity of 10 requests per msec here.
	 */
	if (auth->ds_jiffies == jiffies) {
		auth->ds_time[1] += 100;
	} else {
		p[0] = htonl(jiffies*HZ/1000000);
		p[1] = htonl(jiffies*HZ%1000000);
		auth->ds_jiffies = jiffies;
		auth->ds_time[0] = auth->ds_jiffies / (100000/HZ);
		auth->ds_time[1] = auth->ds_jiffies * HZ % 100000;
	}

	/* Keep the time stamp for des_validate later */
	req->rq_creddata[0] = auth->ds_time[0];
	req->rq_creddata[1] = auth->ds_time[1];

	/* Encrypt time stamp, and window if necessary */
	cryptbuf[0] = htonl(auth->ds_time[0]);
	cryptbuf[1] = htonl(auth->ds_time[1]);
	if (cred->dc_namekind == ADN_FULLNAME) {
		cryptbuf[2] = htonl(auth->ds_window);
		cryptbuf[3] = htonl(auth->ds_window-1);
		status = des_cbc_encrypt((des_cblock *) cryptbuf,
					(des_cblock *) cryptbuf, 16,
					cred->dc_keysched, 
					(des_cblock *) &ivec,
					DES_ENCRYPT);
	} else {
		cryptbuf[2] = cryptbuf[3] = 0;
		status = des_ecb_encrypt((des_cblock *) cryptbuf,
					(des_cblock *) cryptbuf,
					cred->dc_keysched, DES_ENCRYPT);
	}
	if (status) {
		printk("authdes_marshal: DES encryption failed (status %d)\n",
				status);
		return NULL;
	}

	/* Now XDR the lot */
	*p++ = htonl(RPC_AUTH_DES);
	hold = p++;
	*p++ = htonl(cred->dc_namekind);

	if (cred->dc_namekind == ADN_FULLNAME) {
		int	namlen;

		*p++ = 0;
		namlen = sprintf((char *) (p+1), "unix.%d@%s",
				cred->dc_uid, system_utsname.domainname);
		*p++ = htonl(namlen);
		p += ((namlen + 3) >> 2);
		*p++ = cred->dc_xkey[0];	/* encrypted session key */
		*p++ = cred->dc_xkey[1];
		*p++ = cryptbuf[2];		/* encrypted window */
	} else {
		*p++ = cred->dc_nickname;
	}

	*hold = htonl(p - hold - 1);

	/* Now do the verifier */
	*p++ = htonl(RPC_AUTH_DES);
	*p++ = htonl(12);
	*p++ = cryptbuf[0];			/* encrypted time stamp */
	*p++ = cryptbuf[1];
	*p++ = cryptbuf[3];			/* encrypted window - 1 */

	task->tk_suid_retry = 0;

	return p;
}

/*
 * Update credentials
 */
static int
des_refresh(struct rpc_task *task)
{
	struct des_cred	*cred = (struct des_cred *) task->tk_cred,
			*newcred;

	dprintk("RPC: %4d authdes_refresh (%d)\n", task->tk_pid, cred->dc_uid);
	task->tk_status = 0;

	if (cred->dc_flags & RPCAUTH_CRED_UPTODATE)
		return task->tk_status;

	/* If some other process is already talking to keyserv, don't
	 * interfere */
	if (cred->dc_flags & RPCAUTH_CRED_LOCKED) {
		task->tk_timeout = 60 * HZ;
		task->tk_callback = des_refresh_done;
		rpc_sleep_on(&cred->dc_wait, task, NULL, 0);
		return task->tk_status;
	}

	if (cred->dc_nxtrfr >= jiffies) {
		cred->dc_flags |= RPCAUTH_CRED_UPTODATE;
		return task->tk_status;
	}

	if (cred->dc_count == 1) {
		cred->dc_flags |= RPCAUTH_CRED_LOCKED;
	} else {
		/* Remove cred from credcache, and create a new handle. First,
		 * insert the (locked) new handle into the cache, then release
		 * the old one (marking it for immediate destruction). */
		if (!(newcred = des_new_cred(task, cred->dc_uid)))
			return task->tk_status = -EAGAIN;
		newcred->dc_flags |= RPCAUTH_CRED_LOCKED;
		newcred->dc_uid    = cred->dc_uid;
		rpcauth_insert_credcache(task->tk_auth,
					(struct rpc_cred *) newcred);
		cred->dc_flags    |= RPCAUTH_CRED_DEAD;
		rpcauth_releasecred(task);
		task->tk_cred      = (struct rpc_cred *) cred = newcred;
	}

	/* Generate a new key and have keyserv encrypt it. */
	des_generate_key(&cred->dc_key);
	des_set_key(&cred->dc_key, cred->dc_keysched);
	cred->dc_cryptarg.remotename = task->tk_client->cl_server;
	cred->dc_cryptarg.key = &cred->dc_key;
	cred->dc_cryptres.key = &cred->dc_xkey;

	/* This callback will be invoked when the child terminates */
	task->tk_callback = des_refresh_done;

	keycall_cryptsession(task, 1, cred->dc_uid,
					&cred->dc_cryptarg,
					&cred->dc_cryptres);

	return task->tk_status;
}

/*
 * keycall complete. Wake up everyone waiting on the credential.
 * It doesn't matter if the call was successful or not, because
 * all waiting tasks will now call call_reserve which makes sure
 * that the handle is uptodate.
 */
static void
des_refresh_done(struct rpc_task *task)
{
	struct des_cred	*cred = (struct des_cred *) task->tk_cred;

	if (task->tk_status >= 0)
		cred->dc_flags |= RPCAUTH_CRED_UPTODATE;
	cred->dc_flags &= ~RPCAUTH_CRED_LOCKED;
	rpc_wake_up(&cred->dc_wait);
}

static u32 *
des_validate(struct rpc_task *task, u32 *p)
{
	struct des_cred	*cred = (struct des_cred *) task->tk_cred;
	struct rpc_rqst	*req = task->tk_rqstp;
	u32		timestamp[2];
	int		status;

	if (ntohl(*p++) != RPC_AUTH_DES)
		return NULL;
	if (ntohl(*p++) != 4*DES_READSLACK)
		return NULL;

	status = des_ecb_encrypt((des_cblock *) p,
				 (des_cblock *) &timestamp,
				 cred->dc_keysched,
				 DES_DECRYPT);
	p += 2;
	if (status) {
		printk(KERN_NOTICE
			"authdes_validate: DES decryption failed!\n");
		return NULL;
	}

	if (ntohl(timestamp[0]) != req->rq_creddata[0] - 1
	 || ntohl(timestamp[1]) != req->rq_creddata[1]) {
		printk(KERN_NOTICE "authdes: bad verifier from %s\n",
			task->tk_client->cl_server);
		return NULL;
	}

	cred->dc_namekind = ADN_NICKNAME;
	cred->dc_nickname = *p++;

	return p;
}

struct rpc_authops	authdes_ops = {
	RPC_AUTH_UNIX,
#ifdef RPC_DEBUG
	"DES",
#endif
	des_create,
	des_destroy,
	des_create_cred,
	des_destroy_cred,
	des_match_cred,
	des_marshal,
	des_refresh,
	des_validate
};

#ifdef MODULE
int
init_module(void)
{
	return rpcauth_register(&authdes_ops);
}

void
cleanup_module(void)
{
	rpcauth_unregister(&authdes_ops);
}
#endif
