/*
 * linux/net/sunrpc/keycall.c
 *
 * RPC keyserver client implementation.
 *
 * Copyright (C) 1996, Alexander N. Kuznetsiov <kuznet@inr.ac.ru>
 *		       Olaf Kirch <okir@monad.swb.de>
 */

#include <linux/types.h>
#include <linux/malloc.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/utsname.h>
#include <linux/sunrpc/clnt.h>
#include <linux/sunrpc/keycall.h>

#define KEY_PROGRAM		100029
#define KEY_VERSION		1

#define KEYCALL_NULL		0
#define KEYCALL_SETKEY		1
#define KEYCALL_ENCRYPT		2
#define KEYCALL_DECRYPT		3
#define KEYCALL_GENKEY		4
#define KEYCALL_GETCRED		5

extern struct rpc_program	keyserv_program;

/*
 * Global keyserver client handle.
 */
static struct rpc_clnt	*keyserv_clnt = NULL;

/*
 * Set up a fake UNIX cred (auth_unix.c)
 */
struct rpc_cred *	authunix_fake_cred(struct rpc_task *, uid_t, gid_t);

/*
 * Create client for rpc.keyserv
 */
static struct rpc_clnt *
keyserv_getclient(void)
{
	if (!keyserv_clnt) {
		struct sockaddr_in	sin;
		struct rpc_xprt		*xprt;
		struct rpc_clnt		*clnt;

		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
		sin.sin_port = 0;
		if (!(xprt = xprt_create_proto(IPPROTO_TCP, &sin, NULL)))
			return NULL;
		clnt = rpc_create_client(xprt, "localhost",
				&keyserv_program,
				KEY_VERSION, RPC_AUTH_UNIX);
		if (!clnt) {
			xprt_destroy(xprt);
			return NULL;
		}
		if (keyserv_clnt)
			rpc_destroy_client(clnt);
		else
			keyserv_clnt = clnt;
		keyserv_clnt->cl_softrtry = 0;
		keyserv_clnt->cl_chatty   = 1;
		keyserv_clnt->cl_autobind = 1;
	}

	return keyserv_clnt;
}

void
keycall_cryptsession(struct rpc_task *task, int encrypt, uid_t uid,
					struct cryptkeyarg *argp,
					struct cryptkeyres *resp)
{
	struct rpc_clnt	*kclnt;
	struct rpc_task	*child;

	task->tk_status = 0;
	if (!(kclnt = keyserv_getclient()))
		goto bailout;

	if (!(child = rpc_new_child(kclnt, task)))
		goto bailout;

	/* Fake UNIX creds for this keycall */
	if (!authunix_fake_cred(child, uid, -2)) {
		rpc_release_task(child);
		goto bailout;
	}

	/* Setup call info */
	rpc_call_setup(child, encrypt? KEYCALL_ENCRYPT : KEYCALL_DECRYPT,
						argp, resp, 0);

	/* ... and run the task */
	rpc_run_child(task, child, NULL);
	return;

bailout:
	task->tk_status = -EAGAIN;
	return;
}

static inline int
adjust_iovec(struct iovec *iov, u32 *p)
{
	return iov->iov_len = ((u8 *) p - (u8 *) iov->iov_base);
}

static int
xdr_encode_cryptkeyarg(struct rpc_rqst *req, u32 *p, struct cryptkeyarg *argp)
{
	int	len = strlen(argp->remotename);

	*p++ = htonl(len);
	memcpy(p, argp->remotename, len);
	p += XDR_QUADLEN(len);
	memcpy(p, argp->key, 8);
	p += 2;

	req->rq_slen = adjust_iovec(req->rq_svec, p);
	return 0;
}

static int
xdr_decode_cryptkeyres(struct rpc_rqst *req, u32 *p, struct cryptkeyres *resp)
{
	if (ntohl(*p++) != 0)
		return -EACCES;
	memcpy(resp->key, p, 8);
	return 0;
}

static struct rpc_procinfo	key_procedures[6] = {
      {	"key_null",		/* never needed in kernel-land */
	NULL,
	NULL,
	0, 0
      },
      { "key_set",		/* never needed in kernel-land */
	NULL,
	NULL,
	0, 0
      },
      { "key_encrypt",
	(kxdrproc_t) xdr_encode_cryptkeyarg,
	(kxdrproc_t) xdr_decode_cryptkeyres,
	4+RPC_MAXNETNAMELEN+8, 0
      },
      { "key_decrypt",
	(kxdrproc_t) xdr_encode_cryptkeyarg,
	(kxdrproc_t) xdr_decode_cryptkeyres,
	4+RPC_MAXNETNAMELEN+8, 0
      },
      { "key_gen",		/* not implemented for now */
	NULL,
	NULL,
	0, 0
      },
      { "key_getcred",		/* not implemented for now */
	NULL,
	NULL,
	0, 0
      }
};

static struct rpc_version	keyserv_version1 = {
	1, 5, key_procedures
};

static struct rpc_version *	keyserv_version[] = {
	NULL,
	&keyserv_version1
};

static struct rpc_stat		keyserv_stats;

struct rpc_program	keyserv_program = {
	"keyserv",
	KEY_PROGRAM,
	sizeof(keyserv_version)/sizeof(keyserv_version[0]),
	keyserv_version,
	&keyserv_stats,
};
