/*
 * utils/exportfs/exportfs.c
 *
 * Export file systems to knfsd
 *
 * Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
 */

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <getopt.h>
#include <netdb.h>
#include "xmalloc.h"
#include "nfslib.h"
#include "exportfs.h"
#include "xmalloc.h"
#include "xlog.h"

static void	exportall(int verbose);
static void	unexportall(int verbose);
static void	exportfs(char *arg, char *options, int verbose);
static void	unexportfs(char *arg, int verbose);
static void	dump(int verbose);
static void	error(nfs_export *exp, int err);
static void	usage(void);

int
main(int argc, char **argv)
{
	char	*options = NULL;
	int	f_export = 1;
	int	f_all = 0;
	int	f_verbose = 0;
	int	f_ignore = 0;
	int	i, c;

	xlog_open("exportfs");

	while ((c = getopt(argc, argv, "aio:uv")) != EOF) {
		switch(c) {
		case 'a':
			f_all = 1;
			break;
		case 'i':
			f_ignore = 1;
			break;
		case 'o':
			options = optarg;
			break;
		case 'u':
			f_export = 0;
			break;
		case 'v':
			f_verbose = 1;
			break;
		default:
			usage();
			break;
		}
	}

	if (!f_ignore) {
		export_read(_PATH_EXPORTS);
		xtab_read();
	}

	if (optind == argc) {
		if (f_all) {
			if (f_export)
				exportall(f_verbose);
			else
				unexportall(f_verbose);
		} else {
			dump(f_verbose);
		}
	} else {
		for (i = optind; i < argc; i++) {
			if (f_export)
				exportfs(argv[i], options, f_verbose);
			else
				unexportfs(argv[i], f_verbose);
		}
	}
	return 0;
}

static void
exportall(int verbose)
{
	nfs_export	*exp;
	int		i;

	for (i = 0; i < MCL_MAXTYPES; i++) {
		for (exp = exportlist[i]; exp; exp = exp->m_next) {
			if (verbose)
				printf("exporting %s:%s\n",
					exp->m_client->m_hostname, 
					exp->m_export.e_path);
			if (exp->m_client->m_type == MCL_FQDN) {
				if (!export_export(exp))
					error(exp, errno);
			}
			exp->m_xtabent = 0;	/* force update */
			exp->m_addxtab = 1;
		}
	}
	xtab_write(0);	/* replace file */
}

static void
unexportall(int verbose)
{
	nfs_export	*exp;
	int		i;

	for (i = 0; i < MCL_MAXTYPES; i++) {
		for (exp = exportlist[i]; exp; exp = exp->m_next) {
			if (verbose)
				printf("unexporting %s:%s\n",
					exp->m_client->m_hostname, 
					exp->m_export.e_path);
			if (exp->m_client->m_type == MCL_FQDN) {
				if (!export_unexport(exp))
					error(exp, errno);
			}
			exp->m_xtabent = 0;
			exp->m_addxtab = 0;
		}
	}
	xtab_write(0);	/* clear xtab */
}

static void
exportfs(char *arg, char *options, int verbose)
{
	struct exportent *eep;
	nfs_export	*exp;
	struct hostent	*hp = NULL;
	char		*path;
	char		*hname = arg;
	int		htype;

	if ((path = strchr(arg, ':')) != NULL)
		*path++ = '\0';
	else
		path = "";

	if ((htype = client_gettype(hname)) == MCL_FQDN &&
	    (hp = gethostbyname(hname)) != NULL) {
		exp = export_find(hp, path);
	} else {
		exp = export_lookup(hname, path);
	}

	if (!exp) {
		if (!(eep = mkexportent(hname, path, options)) ||
		    !(exp = export_create(eep)))
			return;
	} else if (!updateexportent(&exp->m_export, options)) {
		return;
	}

	if (verbose)
		printf("exporting %s:%s\n", exp->m_client->m_hostname, 
			exp->m_export.e_path);
	if (exp->m_client->m_type == MCL_FQDN) {
		if (!export_export(exp))
			error(exp, errno);
	}
	exp->m_addxtab = 1;

	xtab_write(0);
}

static void
unexportfs(char *arg, int verbose)
{
	nfs_export	*exp;
	struct hostent	*hp;
	char		*path;
	char		*hname = arg;
	int		htype;

	if ((path = strchr(arg, ':')) != NULL)
		*path++ = '\0';

	if ((htype = client_gettype(hname)) == MCL_FQDN) {
		hp = gethostbyname(hname);
		hname = (char *) hp->h_name;
	}

	for (exp = exportlist[htype]; exp; exp = exp->m_next) {
		if (path && strcmp(path, exp->m_export.e_path))
			continue;
		if (htype != exp->m_client->m_type ||
		    strcmp(exp->m_export.e_hostname, hname))
			continue;
		if (verbose)
			printf("unexporting %s:%s\n",
				exp->m_client->m_hostname, 
				exp->m_export.e_path);
		if (exp->m_client->m_type == MCL_FQDN) {
			if (!export_unexport(exp))
				error(exp, errno);
		}
		exp->m_xtabent = 0;
		exp->m_addxtab = 0;
	}

	xtab_write(0);
}

static char
dumpopt(char c, char *fmt, ...)
{
	va_list	ap;

	va_start(ap, fmt);
	printf("%c", c);
	vprintf(fmt, ap);
	va_end(ap);
	return ',';
}

static void
dump(int verbose)
{
	nfs_export	*exp;
	struct exportent *ep;
	int		htype;
	char		*hname, c;

	for (htype = 0; htype < MCL_MAXTYPES; htype++) {
		for (exp = exportlist[htype]; exp; exp = exp->m_next) {
			ep = &exp->m_export;
			if (htype == MCL_ANONYMOUS)
				hname = "<world>";
			else
				hname = ep->e_hostname;
			if (strlen(ep->e_path) > 14)
				printf("%-14s\n\t\t%s", ep->e_path, hname);
			else
				printf("%-14s\t%s", ep->e_path, hname);
			if (!verbose) {
				printf("\n");
				continue;
			}
			c = '(';
			if (ep->e_flags & NFSEXP_READONLY)
				c = dumpopt(c, "ro");
			else
				c = dumpopt(c, "rw");
			if (ep->e_flags & NFSEXP_ASYNC)
				c = dumpopt(c, "async");
			if (ep->e_flags & NFSEXP_GATHERED_WRITES)
				c = dumpopt(c, "wdelay");
			if (ep->e_flags & NFSEXP_INSECURE_PORT)
				c = dumpopt(c, "insecure");
			if (ep->e_flags & NFSEXP_ROOTSQUASH)
				c = dumpopt(c, "root_squash");
			else
				c = dumpopt(c, "no_root_squash");
			if (ep->e_flags & NFSEXP_ALLSQUASH)
				c = dumpopt(c, "all_squash");
			if (ep->e_maptype == CLE_MAP_UGIDD)
				c = dumpopt(c, "mapping=ugidd");
			else if (ep->e_maptype == CLE_MAP_FILE)
				c = dumpopt(c, "mapping=file");
			if (ep->e_anonuid != -2)
				c = dumpopt(c, "anonuid=%d", ep->e_anonuid);
			if (ep->e_anongid != -2)
				c = dumpopt(c, "anongid=%d", ep->e_anongid);

			printf("%c\n", (c != '(')? ')' : ' ');
		}
	}
}

static void
error(nfs_export *exp, int err)
{
	fprintf(stderr, "%s:%s: %s\n", exp->m_client->m_hostname, 
		exp->m_export.e_path, strerror(err));
}

static void
usage(void)
{
	fprintf(stderr, "usage: exportfs [-a] [host:/path]\n");
	exit(1);
}
