/*
 * utils/mountd/mountd.c
 *
 * Authenticate mount requests and retrieve file handle.
 *
 * Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
 */

#include <signal.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "xmalloc.h"
#include "mountd.h"
#include "getopt.h"
#include "rpcmisc.h"
#include "version.h"


static void	usage(int exitcode);
static exports	get_exportlist(void);

static struct option longopts[] =
{
	{ "foreground", 0, 0, 'F' },
	{ "debug", 1, 0, 'd' },
	{ "help", 0, 0, 'h' },
	{ "exports-file", 1, 0, 'f' },
	{ "version", 0, 0, 'v' },
	{ NULL, 0, 0, 0 }
};

bool_t
mount_null_1_svc(struct svc_req *rqstp, void *argp, void *resp)
{
	return 1;
}

bool_t
mount_mnt_1_svc(struct svc_req *rqstp, dirpath *path, fhstatus *res)
{
	struct sockaddr_in *sin = svc_getcaller(rqstp->rq_xprt);
	struct stat	stb;
	nfs_export	*exp;
	char		rpath[MAXPATHLEN+1];
	char		*p = *path;

	if (*p == '\0')
		p = "/";

	/* Reload /etc/xtab if necessary */
	auth_reload();

	/* Resolve symlinks */
	if (realpath(p, rpath) != NULL) {
		rpath[MAXPATHLEN] = '\0';
		p = rpath;
	}

	/* Now authenticate the intruder... */
	if (!(exp = auth_authenticate("mount", sin, p))) {
		res->fhs_status = NFSERR_ACCES;
		return 1;
	}
	if (stat(p, &stb) < 0) {
		xlog(L_WARNING, "can't stat exported dir %s: %s",
				p, strerror(errno));
		res->fhs_status = NFSERR_NOTDIR;
	} else if (!S_ISDIR(stb.st_mode) && !S_ISREG(stb.st_mode)) {
		xlog(L_WARNING, "%s is not a directory or regular file", p);
		res->fhs_status = NFSERR_NOTDIR;
	} else {
		struct knfs_fh	*fh;

		if (!exp->m_exported)
			export_export(exp);
		if (!exp->m_xtabent)
			xtab_append(exp);

		fh = getfh((struct sockaddr *) sin, stb.st_dev, stb.st_ino);
		if (!fh) {
			xlog(L_WARNING, "getfh failed: %s", strerror(errno));
			res->fhs_status = NFSERR_ACCES;
		} else {
			memcpy(&res->fhstatus_u.fhs_fhandle, fh, 32);
			res->fhs_status = NFS_OK;
			mountlist_add(exp);
		}
	}
	return 1;
}

bool_t
mount_dump_1_svc(struct svc_req *rqstp, void *argp, mountlist *res)
{
	xlog(L_NOTICE, "dump request from %s",
		inet_ntoa(svc_getcaller(rqstp->rq_xprt)->sin_addr));

	*res = mountlist_list();
	return 1;
}

bool_t
mount_umnt_1_svc(struct svc_req *rqstp, dirpath *argp, void *resp)
{
	struct sockaddr_in *sin = svc_getcaller(rqstp->rq_xprt);
	nfs_export	*exp;
	char		*p = *argp;

	if (*p == '\0')
		p = "/";

	if (!(exp = auth_authenticate("unmount", sin, p))) {
		return 1;
	}
	mountlist_del(exp);
	return 1;
}

bool_t
mount_umntall_1_svc(struct svc_req *rqstp, void *argp, void *resp)
{
	/* Reload /etc/xtab if necessary */
	auth_reload();

	mountlist_del_all(svc_getcaller(rqstp->rq_xprt));
	return 1;
}

bool_t
mount_export_1_svc(struct svc_req *rqstp, void *argp, exports *resp)
{
	xlog(L_NOTICE, "export request from %s",
		inet_ntoa(svc_getcaller(rqstp->rq_xprt)->sin_addr));
	*resp = get_exportlist();
	return 1;
}

bool_t
mount_exportall_1_svc(struct svc_req *rqstp, void *argp, exports *resp)
{
	xlog(L_NOTICE, "exportall request from %s",
		inet_ntoa(svc_getcaller(rqstp->rq_xprt)->sin_addr));
	*resp = get_exportlist();
	return 1;
}

static exports
get_exportlist(void)
{
	static exports		elist = NULL;
	struct exportnode	*e, *ne;
	struct groupnode	*g, *ng;
	nfs_export		*exp;
	int			i;

	if (!auth_reload())
		return elist;

	for (e = elist; e != NULL; e = ne) {
		ne = e->ex_next;
		for (g = e->ex_groups; g != NULL; g = ng) {
			ng = g->gr_next;
			xfree(g->gr_name);
			xfree(g);
		}
		xfree(e->ex_dir);
		xfree(e);
	}
	elist = NULL;

	for (i = 0; i < MCL_MAXTYPES; i++) {
		for (exp = exportlist[i]; exp; exp = exp->m_next) {
			for (e = elist; e != NULL; e = e->ex_next) {
				if (!strcmp(exp->m_export.e_path, e->ex_dir))
					break;
			}
			if (!e) {
				e = (struct exportnode *) xmalloc(sizeof(*e));
				e->ex_next = elist;
				e->ex_groups = NULL;
				e->ex_dir = strdup(exp->m_export.e_path);
				elist = e;
			}

			g = (struct groupnode *) xmalloc(sizeof(*g));
			g->gr_name = xstrdup(exp->m_export.e_hostname);
			g->gr_next = e->ex_groups;
			e->ex_groups = g;
		}
	}

	return elist;
}

int
main(int argc, char **argv)
{
	char	*export_file = _PATH_EXPORTS;
	int	foreground = 0;
	int	c;

	/* Initialize logging. */
	xlog_open("mountd");

	rpc_init("mountd", MOUNTPROG, MOUNTVERS, mount_dispatch, 0, 0);

	/* Parse the command line options and arguments. */
	opterr = 0;
	while ((c = getopt_long(argc, argv, "Fd:f:hv", longopts, NULL)) != EOF)
		switch (c) {
		case 'F':
			foreground = 1;
			break;
		case 'd':
			xlog_sconfig(optarg, 1);
			break;
		case 'f':
			export_file = optarg;
			break;
		case 'h':
			usage(0);
			break;
		case 'v':
			printf("kmountd %s\n", VERSION);
			exit(0);
		case 0:
			break;
		case '?':
		default:
			usage(1);
		}

	/* No more arguments allowed. */
	if (optind != argc)
		usage(1);

	auth_init(export_file);

	if (!foreground) {
		/* We first fork off a child. */
		if ((c = fork()) > 0)
			exit(0);
		if (c < 0) {
			xlog(L_FATAL, "mountd: cannot fork: %s\n",
						strerror(errno));
		}
		/* Now we remove ourselves from the foreground. */
		(void) close(0);
		(void) close(1);
		(void) close(2);
		setsid();
		xlog_background();
	}

	svc_run();

	xlog(L_ERROR, "Ack! Gack! svc_run returned!\n");
	exit(1);
}

static void
usage(int n)
{
	fprintf(stderr,
"Usage: rpc.mountd [-Fhnpv] [-d kind] [-f exports-file]\n"
"       [--debug kind] [--help] [--version] [--exports-file=file]\n");
	exit(n);
}
